"""
This module contains implementations of the ``MutableSequence`` ABC using different
data structures.

All public classes from this module implement the ``collections.abc.MutableSequence``
ABC in a way that is intended to be equivalent to the built-in ``list`` type.
"""

from dataclasses import dataclass
from itertools import islice, count, repeat

from abc import abstractmethod, ABC
from typing import (
    Any,
    Optional,
    TypeVar,
    Union,
    Generic,
    Iterator,
    Iterable,
    MutableSequence,
    overload,
)
from collections.abc import Sized

T = TypeVar("T")


def _nth(index: int, iterable: Iterable[T]) -> T:
    """Returns the nth item from an iterable.

    Raises:
       IndexError: if index is incorrect.
    """
    try:
        return next(islice(iterable, index, None))
    except StopIteration:
        raise IndexError("Index out of range")


def _normalize_index(index: int, len_source: Union[int, Sized]) -> int:
    length = len_source if isinstance(len_source, int) else len(len_source)
    translated_index = index if index >= 0 else length - abs(index)

    if translated_index < 0 or translated_index >= length:
        raise IndexError("Index out of range")

    return translated_index


def _normalize_and_clip_index(index: int, len_source: Union[int, Sized]) -> int:
    """Returns index closest to a given index in the sequence of specified length.

    Negative indexing is supported. For indexes greater than the last proper index
    length of the sequence will be returned. If negative indexing would traverse beyond
    the index 0, 0 is returned."""
    length = len_source if isinstance(len_source, int) else len(len_source)
    return slice(index, index + 1).indices(length)[0]


def _indexes_for_slice(slice_: slice, len_source: Union[int, Sized]) -> Iterable[int]:
    """Returns indexes of all elements that would be returned for a given slice.

    Args:
        slice_ (slice): Slice that would be passed to sequence ``s`` of a given length
          in a following manner: ``s[slice_]``
        len_source (Union[int, Sized]): Length of an addressed sequence of a sized
          object to be the source of the same.

    Returns:
        Iterable[int]: All of the indexes for a ``slice_`` in a correct order.
    """
    length = len_source if isinstance(len_source, int) else len(len_source)
    return range(length)[slice_]


def _indexes_for_slice_deletion(
    slice_: slice, len_source: Union[int, Sized]
) -> Iterable[int]:
    """Returns indexes of all elements to delete while removing a given slice.

    Given index is correct in the sequence modified **by removing all of the prior
    indexes**.

    Args:
        slice_ (slice): Slice that would be passed to sequence ``s`` of a given length
          in a following manner: ``del s[slice_]``
        len_source (Union[int, Sized]): Length of an addressed sequence of a sized
          object to be the source of the same.

    Returns:
        Iterable[int]: All of the indexes of elements to remove from ``s`` successively.
    """
    length = len_source if isinstance(len_source, int) else len(len_source)

    offset_source = count() if slice_.step is None or slice_.step > 0 else repeat(0)

    return (
        base_index - offset
        for base_index, offset in zip(range(length)[slice_], offset_source)
    )


@dataclass
class _LinkedListNode(Generic[T]):
    value: T
    next: Optional["_LinkedListNode"] = None


class _ListABC(MutableSequence[T], Generic[T], ABC):
    def __init__(self, values: Optional[Iterable] = None):
        self._head: Optional[_LinkedListNode[T]] = None
        if values is None:
            return

        try:
            iterator = iter(values)
        except TypeError:
            return

        self.extend(iterator)

    @abstractmethod
    def _get_by_index(self, index: int) -> T:
        pass

    @abstractmethod
    def _get_by_slice(self, slice_: slice) -> MutableSequence[T]:
        pass

    @overload
    def __getitem__(self, key: int) -> T:
        pass

    @overload
    def __getitem__(self, key: slice) -> MutableSequence[T]:
        pass

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self._get_by_slice(key)
        else:
            return self._get_by_index(key)

    @abstractmethod
    def _set_by_index(self, index: int, value: T) -> None:
        pass

    @abstractmethod
    def _set_by_slice(self, slice_: slice, values: Iterable[T]) -> None:
        pass

    @overload
    def __setitem__(self, key: int, value: T) -> None:
        ...

    @overload
    def __setitem__(self, key: slice, values: Iterable[T]) -> None:
        ...

    def __setitem__(self, key, value) -> None:
        if isinstance(key, slice) and isinstance(value, Iterable):
            self._set_by_slice(key, value)
        elif isinstance(key, int):
            self._set_by_index(key, value)
        else:
            raise TypeError(
                "Arguments must either of types (int, Any) or (slice, Iterable)"
            )

    def __repr__(self) -> str:
        values_str = ", ".join(repr(e) for e in self)
        return f"{self.__class__.__name__}([{values_str}])"

    @abstractmethod
    def _del_by_index(self, key: int) -> None:
        pass

    @abstractmethod
    def _del_by_slice(self, slice_: slice) -> None:
        pass

    def __delitem__(self, key: Union[int, slice]) -> None:
        if isinstance(key, slice):
            self._del_by_slice(key)
        else:
            self._del_by_index(key)


class _LinkedListAbstractBase(_ListABC, Generic[T], ABC):
    def _get_by_index(self, index: int) -> T:
        return _nth(_normalize_index(index, self), iter(self))

    def _get_by_slice(self, slice_: slice) -> MutableSequence[T]:
        return self.__class__(e.value for e in self._iter_nodes_by_slice(slice_))

    def _set_by_index(self, index: int, value: T) -> None:
        nth_node: _LinkedListNode[T] = _nth(
            _normalize_index(index, self), self._iter_nodes()
        )
        nth_node.value = value

    def _set_by_slice(self, slice_: slice, values: Iterable[T]) -> None:
        values_iter = iter(values)
        for node in self._iter_nodes_by_slice(slice_):
            node.value = next(values_iter)

    def _del_by_slice(self, slice_: slice) -> None:
        for i in _indexes_for_slice_deletion(slice_, self):
            del self[i]

    def _is_empty(self):
        return self._head is None

    def __bool__(self) -> bool:
        return not self._is_empty()

    @abstractmethod
    def _iter_nodes(self) -> Iterator[_LinkedListNode[T]]:
        pass

    # TODO: Improve efficiency by not iterating from the start each time. This affects
    # performance of ``_get_by_slice`` and ``_set_by_slice``.
    def _iter_nodes_by_slice(self, slice_: slice) -> Iterator[_LinkedListNode[T]]:
        return (_nth(i, self._iter_nodes()) for i in _indexes_for_slice(slice_, self))

    def __iter__(self) -> Iterator[T]:
        return (e.value for e in self._iter_nodes())

    def __len__(self) -> int:
        return sum(1 for _ in self)


class LinkedList(_LinkedListAbstractBase, Generic[T]):
    """Equivalent of a built-in ``list`` implemented using **linked list**.

    All of the public methods are intended to work in the same way as ``list``. Current
    implementations of slice-based operations aren't time-efficient.
    """

    def _delete_first(self):
        if self._head is not None:
            self._head = self._head.next

    def _del_by_index(self, key: int) -> None:
        key = _normalize_index(key, self)
        if key == 0:
            self._delete_first()
            return

        prev_node = _nth(key - 1, self._iter_nodes())
        to_remove = prev_node.next
        prev_node.next = to_remove.next if to_remove is not None else None

    def _iter_nodes(self) -> Iterator[_LinkedListNode[T]]:
        current = self._head
        while current is not None:
            yield current
            current = current.next

    def insert(self, index_after_inserted: int, obj: Any):
        index_after_inserted = _normalize_and_clip_index(index_after_inserted, self)
        if index_after_inserted == 0:
            self._insert_first(obj)
            return

        before_inserted = _nth(index_after_inserted - 1, self._iter_nodes())
        to_insert = _LinkedListNode(value=obj, next=before_inserted.next)
        before_inserted.next = to_insert

    def _insert_first(self, obj: Any):
        new_node = _LinkedListNode(value=obj)
        new_node.next = self._head
        self._head = new_node

    def _get_last_node(self) -> _LinkedListNode[T]:
        if self._head is None:
            raise ValueError("This list is empty")

        last = self._head
        while last.next is not None:
            last = last.next
        return last

    def extend(self, iterable: Iterable[T]) -> None:
        iter_ = iter(iterable)
        if self._head is None:
            try:
                self._insert_first(next(iter_))
            except StopIteration:
                return

        last: _LinkedListNode[T] = self._get_last_node()
        for e in iter_:
            last.next = _LinkedListNode(value=e)
            last = last.next



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['noxcollections']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'noxcollections',
    'version': '0.2.0.dev3',
    'description': 'Pure python implementation of simple data structures and some basic algorithms focused on readability.',
    'long_description': "noxcollections\n==============\n\nsimple, clean, pure python collections\n---------------------------------------\n\n**noxcollections** is pure Python implementation of basic data structures and a\nfew basic algorithms operating on them.\n\nDisclaimer\n==========\n\nThis project's main purpose is to get myself acquainted to the process of\ndeveloping a proper python package with the common tooling used for testing,\nlinting, type checking, and building in python.\n\nAdditionally I hope to refresh and better structure my knowledge about the\nimplemented data structures and algorithms.\n\nFor the reasons stated above this project is mainly meant for educational\npurposes. **If you need performance built-in package** ``collections`` **as well\nas some other packages from PyPI will probably serve you better** (due to better\noptimization as well as the fact that most of them are implemented as C\nextensions).\n\nGoals\n=====\n\nFollowing goals are most important to this project:\n\n- Clean, modern code\n- Full test coverage\n- Fully typed public API\n\nCurrent state\n=============\n\nThe following lists contain planned features and their state current state.\n\nThis package is currently before its first release. \n\nData structures\n---------------\n\n- [ ] Linked lists and other data structures that can be based on them\n    - [x] Linked list\n    - [ ] Doubly linked list\n    - [x] Stack\n    - [ ] Queues\n- [ ] Other list types\n- [ ] Tree based structures\n- [ ] Graphs based structures\n\nAlgorithms\n----------\n\n- [x] Search Algorithms\n    - [x] Binary search\n    - [x] Linear search\n- [ ] Sorting Algorithms\n    - [ ] Bubble sort\n    - [ ] Merge sort \n    - [ ] Quick sort\n- [ ] Graph algorithms (TBD)\n\nHow to use\n==========\n\nFor use only Python(``3.8+``) interpreter is needed. To run test/linters/mypy\nuse of Poetry_ is recommended. Further instructions will assume you have it \ninstalled.\n\n.. _Poetry: https://python-poetry.org/\n\nInstallation\n------------\n\nInstall using pip\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nYou can install last release version of this package from PyPI_ using ``pip``.\n\n.. code-block:: bash\n\n    pip install noxcollections\n\n.. _PyPi: https://pypi.org/project/noxcollections/\n\nUsing Poetry\n------------\n\nFirst install the package in the environment manage by Poetry:\n\n.. code-block:: bash\n\n    poetry install  # build, from main repository directory\n\nAnd then call your script from the main repository's directory prefacing it \nwith ``poetry`` for example if your file is int same directory and as called\n``main.py``::\n\n    poetry python main.py\n\nTesting and linting\n===================\n\nTesting uses the pytest_ package for testing and tox_ to run the test suite\nagainst multiple versions of Python:\n\n- ``3.8``\n- ``3.9``\n- ``3.10``\n\nTox configurations also does linting (using black_ and flake8_) and type\nchecking (using mypy_)\n\nTo run all of the aforementioned tools using tox (after installing all of the \ndependencies and package using ``poetry install``)::\n\n    poetry run tox\n\nTo run those tools more selectively check the ``tox.ini`` file to check the names\nof the environments you wish to run.\n\n.. _pytest: https://pytest.org/\n.. _tox: https://tox.wiki/\n.. _black: https://pypi.org/project/black/\n.. _flake8: https://flake8.pycqa.org/\n.. _mypy: http://www.mypy-lang.org/",
    'author': 'Marcin Śladewski',
    'author_email': 'mar4sladewski@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NoxMar/noxcollections',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""This module contains ABC for the stack ADT as well as its implementation based
on a MutableSequence implementation passed by the caller."""

from abc import ABC, abstractmethod

from noxcollections.lists import LinkedList

from typing import (
    Optional,
    Iterable,
    TypeVar,
    Generic,
    Sized,
    Union,
    Any,
    Callable,
    MutableSequence,
)

T = TypeVar("T")
S = TypeVar("S", Any, None)


class StackABC(Generic[T], Sized, ABC):
    """ABC for stack ADT."""

    @abstractmethod
    def __init__(self, iterable: Optional[Iterable[T]] = None) -> None:
        """ABC for stack ADT. If ``iterable`` is passed elements are added to the stack.

        Passing ``iterable`` is equivalent to adding all of the elements from it in
        order.

        Args:
            iterable (Optional[Iterable[T]]): If supplied elements of this will be added
                added to the stack.
        """
        if iterable is None:
            return

        for e in iterable:
            self.push(e)

    @abstractmethod
    def __len__(self) -> int:
        """Returns number of the elements on the stack."""

    def __bool__(self) -> bool:
        return bool(len(self))

    @abstractmethod
    def top(self) -> T:
        """Returns top(last added) element from the stack **without removing it**.

        Raises ``IndexError`` if the stack is empty.

        Returns:
            T: Top-most element from the stack.

        Raises:
            IndexException: If the stack is empty.
        """

    def peek(self, default: S = None) -> Union[T, S]:
        """Returns top element form the stack or ``default`` if the stack is empty.

        Element is not removed. Works the same way as ``top`` but returns a default
        value instead of raising.

        Args:
            default (S): Value to be returned if the stack is empty.

        Returns:
            Union[T, S]: Top-most element from the stack or ``default`` if the stack is
                empty.
        """
        try:
            return self.top()
        except IndexError:
            return default

    @abstractmethod
    def push(self, element: T) -> None:
        """Adds ``element`` as the top-most element of the stack.

        Args:
            element (T): element to be added as the top of the stack.
        """

    @abstractmethod
    def pop(self) -> T:
        """Returns and **removes**  the top-most element form the stack.

        Raises:
            IndexError: if the stack is empty.
        """


class Stack(StackABC, Generic[T]):
    """A stack implementation based on MutableSequence factory passed as a parameter.

    To achieve O(1) time complexity on all operations used MutableSequence
    implementation should ideally have the following operations of time complexity O(1):
    1. Insertion at the first index (``s.insert(0, elem)``)
    2. Deletion at the first index (``del s[0]``)
    3. Length calculation (``len(s)``)

    By default ``noxcollections.lists.LinkedList`` is used as the backing
    ``MutableSequence``. Using ``list`` is not recommended as inserting and deleting
    at the start of the ``lists`` have the time complexity of ``O(n)``.
    """

    def __init__(
        self,
        iterable: Optional[Iterable[T]] = None,
        sequence_factory: Callable[[], MutableSequence] = LinkedList,
    ) -> None:
        """Creates a stack based on a ``MutableSequence`` implementing object.

        If time complexity of operations is important refer to class'es documentation
        for conditions to achieve O(1) time complexity for all operations while choosing
        the type of the sequence returned by the ``sequence_factory``.

        Args:
            iterable (Optional[Iterable[T]], optional): If passed all elements from the
                ``iterable`` are pushed onto the stack after its creation.
            sequence_factory (Callable[[], MutableSequence], optional): Callable
                returning object implementing the MutableSequence protocol. Defaults to
                ``noxcollections.lists.LinkedList``.
        """
        self._sequence_factory = sequence_factory

        self._backing_sequence = sequence_factory()
        super().__init__(iterable)

    def __len__(self) -> int:
        return len(self._backing_sequence)

    def top(self) -> T:
        try:
            return self._backing_sequence[0]
        except IndexError:
            raise IndexError("Stack is empty")

    def push(self, element: T) -> None:
        self._backing_sequence.insert(0, element)

    def pop(self) -> T:
        try:
            return self._backing_sequence.pop(0)
        except IndexError:
            raise IndexError("Stack is empty")

    def __repr__(self) -> str:
        sequence_repr = repr(list(self._backing_sequence)[::-1])
        callable_name = self._sequence_factory.__name__
        return f"{self.__class__.__name__}({sequence_repr}, {callable_name})"
